# Copyright (C) 2021 Unitary Fund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Cirq utility functions."""

from typing import Union

import numpy as np
import cirq


# Note: This is the same as cirq.PauliSumLike but without a typing.ForwardRef
# which causes Sphinx errors.
PauliSumLike = Union[
    int,
    float,
    complex,
    cirq.PauliString,
    cirq.PauliSum,
    cirq.SingleQubitPauliStringGateOperation,
]


def execute(circuit: cirq.Circuit, obs: np.ndarray) -> float:
    """Simulates noiseless wavefunction evolution and returns the
    expectation value of some observable.

    Args:
        circuit: The input Cirq circuit.
        obs: The observable to measure as a NumPy array.

    Returns:
        The expectation value of obs as a float.
    """
    final_wvf = circuit.final_state_vector()
    return np.real(final_wvf.conj().T @ obs @ final_wvf)


def execute_with_shots(
    circuit: cirq.Circuit, obs: PauliSumLike, shots: int
) -> Union[float, complex]:
    """Simulates noiseless wavefunction evolution and returns the
    expectation value of a PauliString observable.

    Args:
        circuit: The input Cirq circuit.
        obs: The observable to measure as a cirq.PauliString.
        shots: The number of measurements.

    Returns:
        The expectation value of obs as a float.
    """

    # Do the sampling
    psum = cirq.PauliSumCollector(circuit, obs, samples_per_term=shots)
    psum.collect(sampler=cirq.Simulator())

    # Return the expectation value
    return psum.estimated_energy()


def execute_with_depolarizing_noise(
    circuit: cirq.Circuit, obs: np.ndarray, noise: float
) -> float:
    """Simulates a circuit with depolarizing noise at level noise.

    Args:
        circuit: The input Cirq circuit.
        obs: The observable to measure as a NumPy array.
        noise: The depolarizing noise as a float, i.e. 0.001 is 0.1% noise.

    Returns:
        The expectation value of obs as a float.
    """
    circuit = circuit.with_noise(cirq.depolarize(p=noise))  # type: ignore
    simulator = cirq.DensityMatrixSimulator()
    rho = simulator.simulate(circuit).final_density_matrix
    expectation = np.real(np.trace(rho @ obs))
    return expectation


def execute_with_shots_and_depolarizing_noise(
    circuit: cirq.Circuit, obs: PauliSumLike, noise: float, shots: int
) -> Union[float, complex]:
    """Simulates a circuit with depolarizing noise at level noise.

    Args:
        circuit: The input Cirq circuit.
        obs: The observable to measure as a NumPy array.
        noise: The depolarizing noise strength as a float (0.001 is 0.1%)
        shots: The number of measurements.

    Returns:
        The expectation value of obs as a float.
    """
    # Add noise
    noisy = circuit.with_noise(cirq.depolarize(p=noise))  # type: ignore

    # Do the sampling
    psum = cirq.PauliSumCollector(noisy, obs, samples_per_term=shots)
    psum.collect(sampler=cirq.DensityMatrixSimulator())

    # Return the expectation value
    return psum.estimated_energy()



# Copyright (C) 2020 Unitary Fund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Defines utilities for efficiently running collections of circuits generated
by error mitigation techniques to compute expectation values."""

from collections import Counter
import inspect
from typing import Any, Callable, Iterable, List, Sequence, Tuple, Union

import numpy as np

from mitiq._typing import QPROGRAM, QuantumResult
from mitiq.interface import convert_from_mitiq, convert_to_mitiq


def generate_collected_executor(
    executor: Callable[[Union[QPROGRAM, Sequence[QPROGRAM]]], Any],
    max_batch_size: int = 75,
    force_run_all: bool = False,
    **kwargs: Any,
) -> Callable[[Any], List[QuantumResult]]:
    """Returns a new executor function which efficiently collects expectation
    values by detecting identical circuits and calling the executor as few
    times as possible.

    Args:
        executor: A serial or batched executor. A serial executor inputs one
            quantum circuit and outputs one expectation value. A batched
            executor inputs a list of quantum circuits and outputs a list of
            expectation values (one for each circuit).

            An executor is detected as batched if and only if it is annotated
            with a return type that is one of the following:

            * Iterable[float]
            * List[float]
            * Sequence[float]
            * Tuple[float]

        max_batch_size: The maximum number of circuits which can be processed
            with one call to the executor. If the executor is serial, this
            argument can be ignored and defaults to one.
        force_run_all: If True, all circuits are run by the executor even if
            some circuits are identical. Otherwise, the set of unique circuits
            is found and only this set is run by the executor.
        kwargs: Keyword arguments to provide to the executor at each call.
    """
    if not callable(executor):
        raise ValueError("Arg `executor` must be callable.")

    def collected(circuits: Any) -> List[QuantumResult]:
        return Executor(executor, max_batch_size).run(
            circuits, force_run_all=force_run_all, **kwargs
        )

    return collected


class Executor:
    """Tool for efficiently scheduling/executing quantum programs and
    collecting the results.
    """

    def __init__(
        self,
        executor: Callable[[Union[QPROGRAM, Sequence[QPROGRAM]]], Any],
        max_batch_size: int = 75,
    ) -> None:
        """Initializes a Collector.

        Args:
            executor: A function which inputs a program and outputs a float,
                or inputs a sequence of programs and outputs a sequence of
                floats.
            max_batch_size: Maximum number of programs that can be sent in a
                single batch (if the executor is batched).
        """
        self._executor = executor
        self._can_batch = Executor.is_batched_executor(executor)
        self._max_batch_size = max_batch_size

        self._executed_circuits: List[QPROGRAM] = []
        self._computed_results: List[QuantumResult] = []

        self._calls_to_executor: int = 0

    @property
    def can_batch(self) -> bool:
        return self._can_batch

    @property
    def calls_to_executor(self) -> int:
        return self._calls_to_executor

    def run(
        self,
        circuits: Sequence[QPROGRAM],
        force_run_all: bool = False,
        **kwargs: Any,
    ) -> List[QuantumResult]:
        """Runs all input circuits using the least number of possible calls to
        the executor.

        Args:
            circuits: Sequence of circuits to execute using the executor.
            force_run_all: If True, force every circuit in the input sequence
            to be executed (if some are identical). Else, detects identical
            circuits and runs a minimal set.
        """
        if force_run_all:
            to_run = circuits
        else:
            # Make circuits hashable.
            # Note: Assumes all circuits are the same type.
            _, conversion_type = convert_to_mitiq(circuits[0])
            hashable_circuits = [
                convert_to_mitiq(circ)[0].freeze() for circ in circuits
            ]

            # Get the unique circuits and counts
            collection = Counter(hashable_circuits)
            to_run = [
                convert_from_mitiq(circ.unfreeze(), conversion_type)
                for circ in collection.keys()
            ]

        if not self._can_batch:
            for circuit in to_run:
                self._call_executor(circuit, **kwargs)

        else:
            stop = len(to_run)
            step = self._max_batch_size
            for i in range(int(np.ceil(stop / step))):
                batch = to_run[i * step : (i + 1) * step]
                self._call_executor(batch, **kwargs)

        # Expand computed results to all results using counts.
        if force_run_all:
            return self._computed_results

        results_dict = dict(zip(collection.keys(), self._computed_results))
        results = [results_dict[key] for key in hashable_circuits]

        return results

    def _call_executor(
        self, to_run: Union[QPROGRAM, Sequence[QPROGRAM]], **kwargs: Any
    ) -> None:
        """Calls the executor on the input circuit(s) to run. Stores the
        executed circuits in ``self._executed_circuits`` and the computed
        results in ``self._computed_results``.

        Args:
            to_run: Circuit(s) to run.
        """
        result = self._executor(to_run, **kwargs)  # type: ignore
        self._calls_to_executor += 1

        if self.can_batch:
            self._computed_results.extend(result)
            self._executed_circuits.extend(to_run)
        else:
            self._computed_results.append(result)
            self._executed_circuits.append(to_run)

    @staticmethod
    def is_batched_executor(
        executor: Callable[[Union[QPROGRAM, Sequence[QPROGRAM]]], Any]
    ) -> bool:
        """Returns True if the input function is recognized as a "batched
        executor", else False.

        The executor is detected as "batched" if and only if it is annotated
        with a return type that is one of the following:

            * ``Iterable[QuantumResult]``
            * ``List[QuantumResult]``
            * ``Sequence[QuantumResult]``
            * ``Tuple[QuantumResult]``

        where a ``QuantumResult`` is a ``float`` or a ``cirq.Result``.
        Otherwise, it is considered "serial".

        Batched executors can run several quantum programs in a single call.
        See below.

        Args:
            executor: A "serial executor" (1) or a "batched executor" (2).

                (1) A function which inputs a single ``QPROGRAM`` and outputs a
                single ``QuantumResult``.
                (2) A function which inputs a list of ``QPROGRAM``s and outputs
                a list of ``QuantumResult``s (one for each ``QPROGRAM``).

        Returns:
            True if the executor is detected as batched, else False.
        """
        executor_annotation = inspect.getfullargspec(executor).annotations

        return executor_annotation.get("return") in (
            BatchedType[T]  # type: ignore[index]
            for BatchedType in [Iterable, List, Sequence, Tuple]
            for T in QuantumResult.__args__  # type: ignore[attr-defined]
        )



# Copyright (C) 2021 Unitary Fund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import cirq

from mitiq import QPROGRAM
from mitiq.interface import accept_any_qprogram_as_input
from mitiq.cdr.execute import MeasurementResult


def random_x_z_circuit(qubits, n_moments, random_state) -> cirq.Circuit:
    angles = np.linspace(0.0, 2 * np.pi, 6)
    oneq_gates = [cirq.ops.rz(a) for a in angles] + [cirq.ops.rx(np.pi / 2)]
    gate_domain = {oneq_gate: 1 for oneq_gate in oneq_gates}

    return cirq.testing.random_circuit(
        qubits=qubits,
        n_moments=n_moments,
        op_density=1.0,
        gate_domain=gate_domain,
        random_state=random_state,
    )


def random_x_z_cnot_circuit(qubits, n_moments, random_state) -> cirq.Circuit:
    angles = np.linspace(0.0, 2 * np.pi, 8)
    oneq_gates = [
        gate(a) for a in angles for gate in (cirq.ops.rx, cirq.ops.rz)
    ]
    gate_domain = {oneq_gate: 1 for oneq_gate in oneq_gates}
    gate_domain.update({cirq.ops.CNOT: 2})

    return cirq.testing.random_circuit(
        qubits=qubits,
        n_moments=n_moments,
        op_density=1.0,
        gate_domain=gate_domain,
        random_state=random_state,
    )


@accept_any_qprogram_as_input
def executor(
    circuit: QPROGRAM, noise_level: float = 0.1, shots: int = 8192
) -> MeasurementResult:
    """Returns computational basis measurements after executing the circuit
    with depolarizing noise.

    Args:
        circuit: Circuit to execute.
        noise_level: Probability of depolarizing noise after each moment.
        shots: Number of samples to take.

    Returns:
        Dictionary where each key is a bitstring (binary int) and each value
        is the number of times that bitstring was measured.
    """
    circuit = circuit.with_noise(cirq.depolarize(p=noise_level))
    circuit.append(cirq.measure(*circuit.all_qubits(), key="z"))

    result = cirq.DensityMatrixSimulator().run(circuit, repetitions=shots)
    return result.histogram(key="z")


def simulator(circuit: QPROGRAM, shots: int = 8192) -> MeasurementResult:
    """Returns computational basis measurements after executing the circuit
    (without noise).

    Args:
        circuit: Circuit to simulate.
        shots: Number of samples to take.

    Returns:
        Dictionary where each key is a bitstring (binary int) and each value
        is the number of times that bitstring was measured.
    """
    return executor(circuit, noise_level=0.0, shots=shots)


@accept_any_qprogram_as_input
def simulator_statevector(circuit: QPROGRAM) -> np.ndarray:
    """Returns the final wavefunction (as a numpy array) of the circuit.

    Args:
        circuit: Circuit to simulate.
    """
    return cirq.Simulator().simulate(circuit).final_state_vector



# Copyright (C) 2021 Unitary Fund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""API for using Clifford Data Regression (CDR) error mitigation."""

from functools import wraps
from typing import Any, Callable, cast, List, Optional, Sequence, Tuple, Union

import numpy as np
from scipy.optimize import curve_fit

from cirq import Circuit

from mitiq.interface import accept_any_qprogram_as_input

from mitiq.cdr import (
    generate_training_circuits,
    linear_fit_function,
    linear_fit_function_no_intercept,
)
from mitiq.cdr.execute import calculate_observable, MeasurementResult
from mitiq.zne.scaling import fold_gates_at_random


@wraps(accept_any_qprogram_as_input)
def execute_with_cdr(
    circuit: Circuit,
    executor: Callable[[Circuit], MeasurementResult],
    simulator: Callable[[Circuit], Union[MeasurementResult, np.ndarray]],
    observables: List[np.ndarray],
    num_training_circuits: int = 10,
    fraction_non_clifford: float = 0.1,
    fit_function: Callable[..., float] = linear_fit_function,
    num_fit_parameters: Optional[int] = None,
    scale_factors: Sequence[float] = (1,),
    scale_noise: Callable[[Circuit, float], Circuit] = fold_gates_at_random,
    full_output: bool = False,
    **kwargs: Any,
) -> Union[List[float], Tuple[List[float], np.ndarray]]:
    """Function for the calculation of an observable from some circuit of
    interest to be mitigated with CDR (or vnCDR) based on [Czarnik2020]_ and
    [Lowe2020]_.

    The circuit of interest must be compiled in the native basis of the IBM
    quantum computers, that is {Rz, sqrt(X), CNOT}, or such that all the
    non-Clifford gates are contained in the Rz rotations.

    The observable/s to be calculated should be input as an array or a list of
    arrays representing the diagonal of the observables to be measured. Note
    these observables MUST be diagonal in z-basis measurements corresponding to
    the circuit of interest.

    Returns mitigated observables list of raw observables (at noise scale
    factors).

    This function returns the mitigated observable/s.
    Args:
        circuit: Circuit of interest compiled in the correct basis.
        executor: User defined function taking a cirq Circuit object and
                  returning a dictionary of counts.
        simulator: User defined function taking a cirq Circuit object and
                   returning either a simulated dictionary of counts or an
                   np.ndarray representing the state vector.
        observables: List of arrays containing the diagonal elements of
                    observable/s of interest to be mitigated. If a list is
                    passed all these observables will be mitigates with the
                    same training set.
        num_training_circuits: Number of training circuits to be used in the
                               mitigation.
        fraction_non_clifford: The fraction of non-Clifford gates to be
                               substituted in the training circuits. The higher
                               this fraction the more costly the simulations,
                               but more successful the mitigation.
        fit_function: The function to map noisy to exact data. Takes array of
                      noisy and data and parameters returning a float.
        num_fit_parameters: The number of parameters the fit_function takes.
        scale_noise: Optional argument containing a user defined function on
                     how to increase the noise. If this argument is given then
                     the mitigation method will be vnCDR.
        scale_factors: Factors by which to scale the noise, should not
                               include 1 as this is just the original circuit.
        full_output: If True, returns ⟨𝛹| O |𝛹⟩ for each training circuit
            |𝛹⟩ (including the original circuit) and each observable O. The
            ith row of this 2d array corresponds to the ith observable, and
            the jth column corresponds to the jth training circuit.
        kwargs: Available keyword arguments are:
            - method_select (string): Specifies the method used to select the
                non-Clifford gates to replace when constructing the
                near-Clifford training circuits. Can be 'uniform' or
                'gaussian'.
            - method_replace (string): Specifies the method used to replace
                the selected non-Clifford gates with a Clifford when
                constructing the near-Clifford training circuits. Can be
                'uniform', 'gaussian', or 'closest'.
            - sigma_select (float): Width of the Gaussian distribution used for
                ``method_select='gaussian'``.
            - sigma_replace (float): Width of the Gaussian distribution used
                for ``method_replace='gaussian'``.
            - random_state (int): Seed for sampling.

    Returns: A list of error mitigated observable expectation values. If
        full_output is True, also returns the observable expectation values
        for each training circuit (see above).

    .. [Czarnik2020] : Piotr Czarnik, Andrew Arramsmith, Patrick Coles,
        Lukasz Cincio, "Error mitigation with Clifford quantum circuit
        data," (https://arxiv.org/abs/2005.10189).
    .. [Lowe2020] : Angus Lowe, Max Hunter Gordon, Piotr Czarnik,
        Andrew Arramsmith, Patrick Coles, Lukasz Cincio,
        "Unified approach to data-driven error mitigation,"
        (https://arxiv.org/abs/2011.01157).
    """
    # Handle keyword arguments for generating training circuits.
    method_select = kwargs.get("method_select", "uniform")
    method_replace = kwargs.get("method_replace", "closest")
    random_state = kwargs.get("random_state", None)
    kwargs_for_training_set_generation = {
        "sigma_select": kwargs.get("sigma_select"),
        "sigma_replace": kwargs.get("sigma_replace"),
    }

    if num_fit_parameters is None and fit_function not in (
        linear_fit_function,
        linear_fit_function_no_intercept,
    ):
        raise ValueError(
            "Must provide arg `num_fit_parameters` for custom fit function."
        )
    num_fit_parameters = cast(int, num_fit_parameters)

    # Generate training circuits.
    training_circuits = generate_training_circuits(
        circuit,
        num_training_circuits,
        fraction_non_clifford,
        method_select,
        method_replace,
        random_state,
        kwargs=kwargs_for_training_set_generation,
    )

    # [Optionally] Scale noise in circuits.
    all_circuits = [
        [scale_noise(c, s) for s in scale_factors]
        for c in [circuit] + training_circuits  # type: ignore
    ]

    # Execute all circuits to get MeasurementResult's. TODO: Allow batching.
    noisy_counts = np.array(
        [[executor(circ) for circ in circuits] for circuits in all_circuits]
    )
    ideal_counts = np.array([simulator(circ) for circ in all_circuits[0]])

    results_dict_circuit_of_interest = noisy_counts[:, 0]

    mitigated_observables = []
    raw_observables = []
    for obs in observables:
        circuit_data = np.array(
            [
                calculate_observable(measurements, obs)
                for measurements in results_dict_circuit_of_interest
            ]
        )
        raw_observables.append(list(circuit_data))

        # Get the noisy ⟨𝛹| O |𝛹⟩ from the noisy (executor) counts.
        noisy_expectation_values = np.array(
            [
                [
                    calculate_observable(measurements, obs)
                    for measurements in row
                ]
                for row in noisy_counts
            ]
        )

        # Get the exact ⟨𝛹| O |𝛹⟩ from the exact (simulator) counts.
        ideal_expectation_values = np.array(
            [
                calculate_observable(
                    state_or_measurements=measurements, observable=obs
                )
                for measurements in ideal_counts
            ]
        )

        # Do the regression.
        fitted_params, _ = curve_fit(
            lambda x, *params: fit_function(x, params),
            noisy_expectation_values,
            ideal_expectation_values,
            p0=np.zeros(num_fit_parameters),
        )
        mitigated_observables.append(fit_function(circuit_data, fitted_params))

    if not full_output:
        return mitigated_observables
    return mitigated_observables, np.array(raw_observables)



# Copyright (C) 2021 Unitary Fund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import Counter
from typing import Counter as CounterType, Dict, Union

import numpy as np

MeasurementResult = Union[Dict[int, int], CounterType[int]]


def calculate_observable(
    state_or_measurements: Union[MeasurementResult, np.ndarray],
    observable: np.ndarray,
) -> float:
    """Returns (estimate of) ⟨𝛹| O |𝛹⟩ for diagonal observable O and quantum
     state |𝛹⟩.

    Args:
        state_or_measurements: Quantum state to calculate the expectation
            value of the observable in. Can be provided as a wavefunction
            (numpy array) or as a dictionary of counts from sampling the
            wavefunction in the computational basis.
        observable: Observable as a diagonal matrix (one-dimensional numpy
            array).
    """
    nqubits = int(np.log2(len(observable)))

    if isinstance(state_or_measurements, np.ndarray):
        observable_values = [
            observable[i]
            * abs(
                np.conjugate(state_or_measurements[i])
                * state_or_measurements[i]
            )
            for i in range(2 ** nqubits)
        ]
    elif isinstance(state_or_measurements, (dict, Counter)):
        probs = normalize_measurements(state_or_measurements)
        observable_values = [
            observable[i] * probs.get(i, 0.0) for i in range(2 ** nqubits)
        ]
    else:
        raise ValueError(
            f"Provided state has type {type(state_or_measurements)} but must "
            f"be a numpy array, Dict[bin, int], or Counter[int]."
        )

    return sum(np.real(observable_values))


def normalize_measurements(counts: MeasurementResult) -> Dict[int, float]:
    """Normalizes the values of the MeasurementResult to get probabilities.

    Args:
        counts: Dictionary/Counter of measurements. Each key is a binary int
            and each value is an int.
    """
    total_counts = sum(counts.values())
    return {
        bitstring: count / total_counts for bitstring, count in counts.items()
    }



# Copyright (C) 2020 Unitary Fund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Methods for benchmarking zero-noise extrapolation on MaxCut-QAOA."""
from typing import Callable, Iterable, List, Optional, Tuple, Union, cast

import numpy as np
from scipy.optimize import minimize

from cirq import (
    Circuit,
    DensityMatrixSimulator,
    identity_each,
    NamedQubit,
    X,
    ZZ,
    H,
)
from mitiq.zne import execute_with_zne
from mitiq.zne.inference import Factory
from mitiq.benchmarks.utils import noisy_simulation


SIMULATOR = DensityMatrixSimulator()


def _make_noisy_backend(
    noise: float, obs: Union[np.ndarray, float]
) -> Callable[[Circuit], float]:
    """Returns a (noisy) execute function for Cirq circuits.

    Args:
        noise: The level of depolarizing noise.
        obs: The observable that the backend should measure.
    """

    def noisy_backend(circ: Circuit) -> float:
        return noisy_simulation(circ, noise, cast(np.ndarray, obs))

    return noisy_backend


def make_maxcut(
    graph: List[Tuple[int, int]],
    noise: float = 0,
    scale_noise: Optional[Callable[[Circuit, float], Circuit]] = None,
    factory: Optional[Factory] = None,
) -> Tuple[
    Callable[[np.ndarray], Optional[float]],
    Callable[[np.ndarray], Circuit],
    float,
]:
    """Makes an executor that evaluates the QAOA ansatz at a given beta
    and gamma parameters.

    Args:
        graph: The MAXCUT graph as a list of edges with integer labelled nodes.
        noise: The level of depolarizing noise.
        scale_noise: The noise scaling method for ZNE.
        factory: The factory to use for ZNE.

    Returns:
        (ansatz_eval, ansatz_maker, cost_obs) as a triple where

        * **ansatz_eval** -- function that evalutes the maxcut ansatz on the
            noisy cirq backend.
        * **ansatz_maker** -- function that returns an ansatz circuit.
        * **cost_obs** -- the cost observable as a dense matrix.

    """
    # get the list of unique nodes from the list of edges
    nodes = list({node for edge in graph for node in edge})
    nodes = list(range(max(nodes) + 1))

    # one qubit per node
    qreg = [NamedQubit(str(nn)) for nn in nodes]

    def cost_step(beta: float) -> Circuit:
        return Circuit(ZZ(qreg[u], qreg[v]) ** beta for u, v in graph)

    def mix_step(gamma: float) -> Circuit:
        return Circuit(X(qq) ** gamma for qq in qreg)

    def qaoa_ansatz(params: np.ndarray) -> Circuit:
        half = int(len(params) / 2)
        betas, gammas = params[:half], params[half:]
        qaoa_steps = sum(
            [
                cost_step(beta) + mix_step(gamma)
                for beta, gamma in zip(betas, gammas)
            ],
            Circuit(),
        )
        return Circuit(H.on_each(qreg)) + qaoa_steps

    # make the cost observable
    identity_matrix = np.eye(2 ** len(nodes))
    cost_mat = -0.5 * sum(
        identity_matrix
        - Circuit([identity_each(*qreg), ZZ(qreg[i], qreg[j])]).unitary()
        for i, j in graph
    )
    noisy_backend = _make_noisy_backend(noise, cost_mat)

    # must have this function signature to work with scipy minimize
    def qaoa_cost(params: np.ndarray) -> Optional[float]:
        qaoa_prog = qaoa_ansatz(params)
        if scale_noise is None and factory is None:
            return noisy_backend(qaoa_prog)
        else:
            assert scale_noise is not None
            return execute_with_zne(
                qaoa_prog,
                executor=noisy_backend,  # type: ignore
                scale_noise=scale_noise,  # type: ignore
                factory=factory,
            )

    return qaoa_cost, qaoa_ansatz, cost_mat


def run_maxcut(
    graph: List[Tuple[int, int]],
    x0: Iterable[float],
    noise: float = 0,
    scale_noise: Optional[Callable[[Circuit, float], Circuit]] = None,
    factory: Optional[Factory] = None,
    optimizer: str = "Nelder-Mead",
    verbose: bool = False,
) -> Tuple[float, np.ndarray, List[Optional[float]]]:
    """Optimizes MaxCut cost function on the given graph using QAOA.

    Args:
        graph: The MAXCUT graph as a list of edges with integer labelled nodes.
        x0: The initial parameters for QAOA [betas, gammas]. The size of x0
            determines the number of steps.
        noise: Depolarizing noise strength.
        scale_noise: The noise scaling method for ZNE.
        factory: The factory to use for ZNE.
        optimizer: Scipy optimization method to use.
        verbose: An option to pass to scipy.minimize.

    Returns:
        A triple of the minimum cost, the values of beta and gamma that
        obtained that cost, and a list of costs at each iteration step.

    Example:
        Run MAXCUT with 2 steps such that betas = [1.0, 1.1] and
        gammas = [1.4, 0.7] on a graph with four edges and four nodes.

        >>> from mitiq.benchmarks.maxcut import run_maxcut
        >>> graph = [(0, 1), (1, 2), (2, 3), (3, 0)]
        >>> x0 = [1.0, 1.1, 1.4, 0.7]
        >>> fun, x, traj = run_maxcut(graph, x0, verbose=True)
        Optimization terminated successfully.
                 Current function value: -4.000000
                 Iterations: 108
                 Function evaluations: 188
    """
    qaoa_cost, *_ = make_maxcut(graph, noise, scale_noise, factory)

    # store the optimization trajectories
    traj = []

    def callback(xk: np.ndarray) -> bool:
        traj.append(qaoa_cost(xk))
        return True

    res = minimize(
        qaoa_cost,
        x0=np.array(x0),
        method=optimizer,
        callback=callback,
        options={"disp": verbose},
    )

    return res.fun, res.x, traj



# Copyright (C) 2020 Unitary Fund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for Collector."""
import pytest
from typing import List

import numpy as np

import cirq
import pyquil

from mitiq.executor.executor import Executor, generate_collected_executor
from mitiq.rem import MeasurementResult


# Serial / batched executors which return floats.
def executor_batched(circuits, **kwargs) -> List[float]:
    return [
        float(v)
        for v in np.full(
            shape=(len(circuits),),
            fill_value=kwargs.setdefault("return_value", 0.0),
        )
    ]


def executor_batched_unique(circuits) -> List[float]:
    return [executor_serial_unique(circuit) for circuit in circuits]


def executor_serial_unique(circuit):
    return float(len(circuit))


def executor_serial_typed(*args, **kwargs) -> float:
    return executor_serial(*args, **kwargs)


def executor_serial(*args, **kwargs):
    return kwargs.setdefault("return_value", 0.0)


def executor_pyquil_batched(programs) -> List[float]:
    for p in programs:
        if not isinstance(p, pyquil.Program):
            raise TypeError

    return [0.0] * len(programs)


# Serial / batched executors which return measurements.
def executor_serial_measurements(circuit) -> MeasurementResult:
    # Assume there is only one measurement key in the circuit.
    assert len(circuit.all_measurement_keys()) == 1

    key = circuit.all_measurement_keys().pop()
    backend = cirq.Simulator()
    return MeasurementResult(
        backend.run(circuit, repetitions=10).measurements[key].tolist()
    )


def executor_batched_measurements(circuits) -> List[MeasurementResult]:
    return [executor_serial_measurements(circuit) for circuit in circuits]


def test_collector_simple():
    collector = Executor(executor=executor_batched, max_batch_size=10)
    assert collector.can_batch
    assert collector._max_batch_size == 10
    assert collector.calls_to_executor == 0


def test_collector_is_batched_executor():
    assert Executor.is_batched_executor(executor_batched)
    assert not Executor.is_batched_executor(executor_serial_typed)
    assert not Executor.is_batched_executor(executor_serial)
    assert not Executor.is_batched_executor(executor_serial_measurements)
    # assert Collector.is_batched_executor(executor_batched_measurements)


@pytest.mark.parametrize("ncircuits", (5, 10, 25))
@pytest.mark.parametrize("executor", (executor_batched, executor_serial))
def test_run_collector_identical_circuits_batched(ncircuits, executor):
    collector = Executor(executor=executor, max_batch_size=10)
    circuits = [cirq.Circuit(cirq.H(cirq.LineQubit(0)))] * ncircuits
    results = collector.run(circuits)

    assert np.allclose(results, np.zeros(ncircuits))
    assert collector.calls_to_executor == 1


@pytest.mark.parametrize("batch_size", (1, 2, 10))
def test_run_collector_nonidentical_pyquil_programs(batch_size):
    collector = Executor(
        executor=executor_pyquil_batched, max_batch_size=batch_size
    )
    assert collector.can_batch

    circuits = [
        pyquil.Program(pyquil.gates.X(0)),
        pyquil.Program(pyquil.gates.H(0)),
    ] * 10
    results = collector.run(circuits)

    assert np.allclose(results, np.zeros(len(circuits)))
    if batch_size == 1:
        assert collector.calls_to_executor == 2
    else:
        assert collector.calls_to_executor == 1


@pytest.mark.parametrize("ncircuits", (10, 11, 23))
@pytest.mark.parametrize("batch_size", (1, 2, 5, 50))
def test_run_collector_all_unique(ncircuits, batch_size):
    collector = Executor(executor=executor_batched, max_batch_size=batch_size)
    assert collector.can_batch

    random_state = np.random.RandomState(seed=1)
    circuits = [
        cirq.testing.random_circuit(
            qubits=4, n_moments=10, op_density=1, random_state=random_state
        )
        for _ in range(ncircuits)
    ]
    results = collector.run(circuits)

    assert np.allclose(results, np.zeros(ncircuits))
    assert collector.calls_to_executor == np.ceil(ncircuits / batch_size)


@pytest.mark.parametrize("ncircuits", (5, 21))
@pytest.mark.parametrize("force_run_all", (True, False))
def test_run_collector_force_run_all_serial_executor_identical_circuits(
    ncircuits, force_run_all
):
    collector = Executor(executor=executor_serial)
    assert not collector.can_batch

    circuits = [cirq.Circuit(cirq.H(cirq.LineQubit(0)))] * ncircuits
    results = collector.run(circuits, force_run_all=force_run_all)

    assert np.allclose(results, np.zeros(ncircuits))
    if force_run_all:
        assert collector.calls_to_executor == ncircuits
    else:
        assert collector.calls_to_executor == 1


@pytest.mark.parametrize("s", (50, 100, 150))
@pytest.mark.parametrize("b", (1, 2, 100))
def test_run_collector_preserves_order(s, b):
    rng = np.random.RandomState(1)

    collector = Executor(executor=executor_batched_unique, max_batch_size=b)
    assert collector.can_batch

    circuits = [
        cirq.Circuit(cirq.H(cirq.LineQubit(0))),
        cirq.Circuit([cirq.H(cirq.LineQubit(0))] * 2),
    ]
    batch = [circuits[i] for i in rng.random_integers(low=0, high=1, size=s)]

    assert np.allclose(collector.run(batch), executor_batched_unique(batch))


@pytest.mark.parametrize("executor", (executor_serial, executor_batched))
@pytest.mark.parametrize("ncircuits", (10, 25))
@pytest.mark.parametrize("rval", (0.0, 1.0))
def test_generate_collected_executor(executor, ncircuits, rval):
    collected_executor = generate_collected_executor(
        executor, return_value=rval
    )
    expvals = collected_executor([cirq.Circuit()] * ncircuits)
    assert np.allclose(expvals, np.full(shape=(ncircuits,), fill_value=rval))


def test_generate_collected_executor_not_callable():
    with pytest.raises(ValueError, match="must be callable"):
        generate_collected_executor(None)



# Copyright (C) 2020 Unitary Fund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Utility functions for benchmarking."""
from typing import cast, Union

import numpy as np

from cirq import (
    Circuit,
    depolarize,
    DensityMatrixSimulator,
    NoiseModel,
    SingleQubitGate,
)

SIMULATOR = DensityMatrixSimulator()


def noisy_simulation(circ: Circuit, noise: float, obs: np.ndarray) -> float:
    """Simulates a circuit with depolarizing noise at level NOISE.

    Args:
        circ: The quantum program as a cirq object.
        noise: The level of depolarizing noise.
        obs: The observable that the backend should measure.

    Returns:
        The observable's expectation value.
    """
    circuit = circ.with_noise(
        cast(Union[None, NoiseModel, SingleQubitGate], depolarize(p=noise))
    )
    result = SIMULATOR.simulate(circuit)
    rho = result.final_density_matrix
    # measure the expectation by taking the trace of the density matrix
    expectation = np.real(np.trace(rho @ obs))
    return expectation
